"""
ROS1 version
"""

import rospy
import rostopic
from threading import Thread
from concurrent.futures import Future
from deprecated import deprecated

import armw.globals
from armw.base_node import BaseNode


class ServiceProxy(object):
    """
    Adds the ability to return a future rather than wait on service call
    """

    def __init__(self, service_name, service_type, persistent, blocking):
        self._service_proxy = rospy.ServiceProxy(
            service_name, service_type, persistent=persistent
        )

        self.service_class = service_type._request_class

        self.blocking = blocking

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            req = args[0]
        else:
            req = self.service_class(**kwargs)

        if self.blocking:
            return self._service_proxy(req)

        else:
            future = Future()

            thread = Thread(
                target=self._wait_for_service_in_thread, args=(req, future)
            )
            thread.daemon = True
            thread.start()

            return future

    def _wait_for_service_in_thread(
        self, req, future
    ) -> None:
        """
        Internal function to be executed in a separate thread.
        Handles service call, fufills promise.

        :param req: request for service
        :future: future to update when we hear something

        :return: none
        """

        try:
            msg = self._service_proxy(req)
            future.set_result(msg)
        except Exception as e:
            future.set_exception(e)


class Ros1Node(BaseNode):
    def __init__(self, node_name):
        try:
            rospy.init_node(node_name)

            armw.globals.NODE_NAME = node_name
            armw.globals.NODE = self
        except rospy.ROSException:
            self.log(2, f"Could not init node, {node_name} are you calling more than once in the same process?")

        super().__init__()

    def log(self, level: int, message: str):
        if level == -1:
            rospy.logdebug(message)
        elif level == 0:
            rospy.loginfo(message)
        elif level == 1:
            rospy.logwarn(message)
        elif level == 2:
            rospy.logerr(message)
        elif level == 3:
            rospy.logfatal(message)

    def log_throttle(self, interval: int, level: int, message: str):
        if level == -1:
            rospy.logdebug_throttle(interval, message)
        elif level == 0:
            rospy.loginfo_throttle(interval, message)
        elif level == 1:
            rospy.logwarn_throttle(interval, message)
        elif level == 2:
            rospy.logerr_throttle(interval, message)

    def get_name(self):
        return rospy.get_name()

    def get_namespace(self):
        return rospy.get_namespace()

    def search_param(self, parameter_name):
        return rospy.search_param(parameter_name)

    def has_param(self, parameter_name):
        return rospy.has_param(parameter_name)

    def get_param(self, parameter_name, default=None):
        """
        Returns the value for a given parameter name, or None if it can't find it
        """
        return rospy.get_param(parameter_name, default)

    def get_param_as_dict(self, parameter_name, default=None):
        """
        This is for ros2, in ros1 it does the same as above function

        Returns the value for a given parameter name, or None if it can't find it
        """
        return self.get_param(parameter_name, default)

    def set_param(self, parameter_name, value):
        rospy.set_param(parameter_name, value)

    def delete_param(self, parameter_name):
        rospy.delete_param(parameter_name)

    def fix_topic_name(self, topic: str) -> str:
        """
        ros2 allows ~ at the start of a topic but requires ~ to be followed by /

        :param topic: topic to splice

        :return: spliced string
        """

        if topic.startswith("~/"):
            return f"~{topic[2:]}"
        return topic

    def get_topic_type(self, topic):
        topic_type, _, _ = rostopic.get_topic_class(self.fix_topic_name(topic))
        return topic_type

    def publish(self, topic: str, message_type, queue_size=10, latch=False):
        return rospy.Publisher(
            self.fix_topic_name(topic), message_type, queue_size=queue_size, latch=latch
        )

    def subscribe(self, topic: str, message_type, callback, queue_size=10, **kwargs):
        if message_type is None:
            from rospy import AnyMsg

            message_type = AnyMsg

        return rospy.Subscriber(self.fix_topic_name(topic), message_type, callback)

    def create_service_server(self, name, service_type, callback):
        return rospy.Service(self.fix_topic_name(name), service_type, callback)

    def create_service_client(self, name: str, service_type, persistent=False, blocking=True):
        return ServiceProxy(
            self.fix_topic_name(name), service_type, persistent=persistent, blocking=blocking
        )

    def wait_for_message(self, topic, message_type, timeout=None, **kwargs):
        if message_type is None:
            from rospy import AnyMsg

            message_type = AnyMsg

        if timeout is None:
            return rospy.wait_for_message(self.fix_topic_name(topic), message_type)
        else:
            return rospy.wait_for_message(
                self.fix_topic_name(topic), message_type, timeout=timeout
            )

    def wait_for_service(self, name, timeout=None):
        if timeout is None:
            rospy.wait_for_service(self.fix_topic_name(name))
        else:
            return rospy.wait_for_service(self.fix_topic_name(name), timeout=timeout)


def spin(*args):
    """
    just spin the node to keep it alive
    """

    rospy.spin()

def init():
    """
    ros1 does not need rclpy.init like ros2
    """

    pass


@deprecated(version='2.0.0', reason="This function is deprecated, please use init instead")
def init_node(node_name):
    rospy.init_node(node_name)


@deprecated(version='2.0.0', reason="This function is deprecated, please use init and spin directly instead")
def run_node(node_object):
    armw.globals.NODE = node_object
    try:
        armw.globals.NODE.spin()
        rospy.loginfo(f"{armw.globals.NODE_NAME} shutdown")
    except rospy.ROSInterruptException:
        rospy.loginfo(f"{armw.globals.NODE_NAME} shutdown (interrupt)")


@deprecated(version='2.0.0', reason="This function is deprecated, please use init and spin directly instead")
def main(node_class: type(BaseNode), node_name: str):
    rospy.init_node(node_name)

    armw.globals.NODE_NAME = node_name
    armw.globals.NODE = node_class()

    try:
        armw.globals.NODE.spin()
        rospy.loginfo(f"{node_name} shutdown")
    except rospy.ROSInterruptException:
        rospy.loginfo(f"{node_name} shutdown (interrupt)")
