"""
This module just uses polymorphism to override the broken
parts of MultiThreadedExecutor
"""

import rclpy
import time

from typing import Union, Optional, Callable

from threading import Lock

from rclpy.executors import (
    ExternalShutdownException,
    ShutdownException,
    TimeoutException,
    ConditionReachedException,
    TimeoutObject,
)


class MultiThreadedExecutor(rclpy.executors.MultiThreadedExecutor):
    """
    rclpy's MultiThreadedExecutor with a lock to protect self._futures.

    overrides the broken _spin_once_impl in rclpy.

    One day if MR #1477 gets backported (or if we no longer care about older ROS2
    versions) this may be removed.
    """

    def __init__(self, num_threads=None, *, context=None) -> None:
        super().__init__(context=context)
        self._futures_lock = Lock()
        self._spinning_lock = Lock()
        self._spinning = False

    def _spin_once_impl(
        self,
        timeout_sec: Optional[Union[float, TimeoutObject]] = None,
        wait_condition: Callable[[], bool] = lambda: False,
    ) -> None:
        try:
            handler, entity, node = self.wait_for_ready_callbacks(
                timeout_sec, None, wait_condition
            )
        except ExternalShutdownException:
            pass
        except ShutdownException:
            pass
        except TimeoutException:
            pass
        except ConditionReachedException:
            pass
        else:
            self._executor.submit(handler)

            with self._futures_lock:
                self._futures.append(handler)
                for future in self._futures:
                    if future.done():
                        self._futures.remove(future)
                        future.result()  # raise any exceptions

    def might_spin_until_future_complete(
        self, future, timeout_sec: Optional[float] = None
    ) -> None:
        """Execute callbacks until a given future is done or a timeout occurs.
        Similair to rclpy's spin until future complete, however it only spins if
        the executor is not already spinning. We do this by maintaining a spinning
        variable in the executor class.

        We might spin if we need to, or we might not!
        """
        self._spinning_lock.acquire()
        if not self._spinning:
            self.spin_until_future_complete(future=future, timeout_sec=timeout_sec)
            self._spinning_lock.release()
        else:
            self._spinning_lock.release()
            timeout_sec = (
                timeout_sec if timeout_sec is not None and timeout_sec >= 0 else 1e9
            )

            start = time.monotonic()
            end = start + timeout_sec
            timeout_left = TimeoutObject(timeout_sec)

            while (
                self._context.ok()
                and not future.done()
                and not future.cancelled()
                and not self._is_shutdown
            ):

                now = time.monotonic()

                # prevent busy waiting
                time.sleep(0.01)
                if now >= end:
                    return

                timeout_left.timeout = end - now
