#!/usr/bin/env python3

"""
ROS2 version

This is not a comprehension solution to converting ROS 1 launch files to ROS 2.
There might well be launch files that this script is not capable of converting.

This script is not designed for normal use. It may be useful if you want an easy
way to share launch files between ros1 and ros2, but in reality you should just
maintain seperate launch files.

The parsing function is recursive, this could be a potential bottleneck for very
large launch files if the depth of the xml tree exceeds the recursion limit.
This is not the most efficient implementation, if is possible to iterate over
the tree non-recursively but, this is the implementation that came to mind first
and its NOT meant for mainstream use.
"""

import xml.etree.ElementTree as ET

import pathlib
import os
import tempfile
import signal
import multiprocessing
import sys
import yaml
from launch import LaunchDescription
from launch.actions import IncludeLaunchDescription
from launch_xml.launch_description_sources import XMLLaunchDescriptionSource
from launch.launch_service import LaunchService

from armw.base_launch import BaseLaunch
import re


class Launch(BaseLaunch):
    """
    ros2 roslaunch implementation
    """

    def __init__(self, launch_file, args=None) -> None:
        self.args = (
            [tuple(arg.split(":=", 1)) for arg in args] if args is not None else None
        )
        self.launch = self.parse_launch(launch_file)
        self.process = None

        # we do not want to fork the proc, we want to spawn a new one for roslaunch.
        multiprocessing.set_start_method("spawn", force=True)

    @staticmethod
    def _launch(launch_file, launch_args):
        """static method to run the LaunchService in a separate thread."""
        try:
            ls = LaunchService()
            ld = LaunchDescription(
                [
                    IncludeLaunchDescription(
                        XMLLaunchDescriptionSource(launch_file),
                        launch_arguments=launch_args,
                    )
                ]
            )

            ls.include_launch_description(ld)
            ls.run()
        except Exception as e:
            print(f"An exception has occured in the launch process: {e}", file=sys.stderr)

    def start(self, blocking: bool = False) -> None:
        """
        starts launch file, if blocking, will block until signalled.

        :param blocking: whether or not to block or return

        :return: none
        """
        if self.process and self.process.is_alive():
            raise ValueError("Launch file is already running")

        self.process = multiprocessing.Process(
            target=Launch._launch, args=(self.launch, self.args), name="ArmwROS2Launch"
        )
        self.process.start()

        if blocking:
            self.process.join()

    def parse_launch(self, launch_file):
        """
        recursive function to parse launch files. Can only go 50 include
        statements deep by default.

        Creates a temp file for the new launch file then converts the syntax to
        ros2
        """

        temp = tempfile.NamedTemporaryFile(delete=False, suffix=".xml")
        tree = ET.parse(launch_file)
        root = tree.getroot()

        path = pathlib.Path(launch_file).parent
        self.convert(root, path)

        tree.write(temp.name, encoding="UTF-8")

        print(f"Converted launch file to: {temp.name}")
        return temp.name

    def shutdown(self):
            """
            Shuts down the launch process and ALL of its child nodes by
            signaling the entire process group.

            Escalate until we are as sure as we can be we killed all the
            procs.
            """
            if not self.process or not self.process.is_alive():
                print("No launch process running to shut down.", file=sys.stderr)
                return

            try:
                pgid = os.getpgid(self.process.pid)

                os.killpg(pgid, signal.SIGINT)

                self.process.join(timeout=20)

                if self.process.is_alive():
                    os.killpg(pgid, signal.SIGKILL)
                    self.process.join(timeout=5)

            except ProcessLookupError:
                print(f"Process {self.process.pid} was already gone.", file=sys.stderr)
            except Exception as e:
                print(f"An error occurred during shutdown: {e}", file=sys.stderr)
            finally:
                self.process = None

    def convert(self, root, path):
        """
        converts ros1 launch file to ros2.
        """
        for child in list(root):
            # $(find) is no longer valid, we need to find all instances and replace.AssertionError
            for key in list(child.attrib.keys()):
                if "$(arg" in child.attrib[key]:
                    child.attrib[key] = re.sub(
                        r"\$\(arg", "$(var", child.attrib[key]
                    )
                if "$(find" in child.attrib[key]:
                    child.attrib[key] = re.sub(
                        r"\$\(find", "$(find-pkg-share", child.attrib[key]
                    )
                if "$(dirname)" in child.attrib[key]:
                    child.attrib[key] = re.sub(
                        r"\$\(dirname\)", str(path), child.attrib[key]
                    )

            if "clear_params" in child.attrib:
                child.attrib.pop("clear_params")

            # If our param contains a / we need to replace it with .
            if child.tag == "param" and "/" in child.attrib["name"]:
                child.attrib["name"] = _convert_param(child.attrib["name"])

            # we need to convert rosparams which are no longer supported
            if child.tag == "rosparam":
                root.remove(child)

                rosparam_content = yaml.safe_load(child.text)

                def create_param_elements(parent_element, data):
                    if isinstance(data, dict):
                        for key, value in data.items():
                            param_name = _convert_param(key)
                            if isinstance(value, (dict, list)):
                                new_param = ET.SubElement(parent_element, "param")
                                new_param.set("name", param_name)
                                create_param_elements(new_param, value)
                            else:
                                new_param = ET.SubElement(parent_element, "param")
                                new_param.set("name", param_name)

                                value_str = str(value)
                                new_param.set("value", value_str)

                    elif isinstance(data, list):
                        # If the list contains dictionaries, expand them as parameters
                        if all(isinstance(item, dict) for item in data):
                            for item_dict in data:
                                create_param_elements(parent_element, item_dict)
                        else:
                            list_as_string = "[" + ", ".join(map(str, data)) + "]"
                            parent_element.set("value", list_as_string)

                # Start the recursive creation from the root of the rosparam content
                # If the rosparam has a 'param' attribute, it becomes the root namespace
                if "param" in child.attrib:
                    namespace_root = ET.Element("param")
                    namespace_root.set("name", child.attrib["param"])
                    root.append(namespace_root)
                    create_param_elements(namespace_root, rosparam_content)
                else:
                    # If no 'param' attribute, directly add the parsed content
                    create_param_elements(root, rosparam_content)

            # ros2 does not like having if / unless in your param, so we will resolve that here
            if "if" in child.attrib or "unless" in child.attrib:
                args = {arg[0]: arg[1] for arg in self.args} if self.args is not None else {}
                # true if if, false if unless, used to compare against condition value
                is_if = "if" in child.attrib
                attribute_name = "if" if is_if else "unless"

                condition_value = _evaluate_conditional(
                    child.attrib[attribute_name], args
                )

                if is_if != condition_value:
                    root.remove(child)
                    continue
                else:
                    child.attrib.pop(attribute_name)

            # env has been swapped to set_env
            if child.tag == "env":
                child.tag = "set_env"

            # check and swap if a node is using namespace
            if child.tag == "node" and "ns" in child.attrib:
                val = child.attrib["ns"]
                child.attrib.pop("ns")
                child.attrib["namespace"] = val

            # exec is used instead of type in node tag
            if child.tag == "node" and "type" in child.attrib:
                val = child.attrib["type"]
                child.attrib.pop("type")
                child.attrib["exec"] = val

            # required is no longer valid
            if child.tag == "node" and "required" in child.attrib:
                child.attrib.pop("required")

            if child.tag == "arg":
                if "doc" in child.attrib:
                    val = child.attrib["doc"]
                    child.attrib.pop("doc")
                    child.attrib["description"] = val

            # check group tag for clear param and namespace
            if child.tag == "group":
                if "ns" in child.attrib:
                    val = child.attrib["ns"]
                    child.attrib.pop("ns")

                    namespace = ET.Element("push-ros-namespace")
                    namespace.set("namespace", val)
                    child.insert(0, namespace)

            # includes need to be converted as well
            if child.tag == "include":
                # we are not going to deal with find-pkg-share as it would require rospack
                file_attrib = child.attrib["file"]

                match = re.search(r"\$\(find-pkg-share (.*?)\)", file_attrib)

                if not match:
                    self.parse_launch(child.attrib["file"])
                else:
                    package_name = match.group(1)
                    print(f"Your launch file contains an include for a launch file in an outside package {package_name}, this file may need to be converted seperately.")

                # In ROS2 we want .xml not .launch as the ext
                if file_attrib.endswith(".launch"):
                    child.attrib["file"] = os.path.splitext(file_attrib)[0] + ".xml"

            # no global params!
            if root.tag != "node" and root.tag != "param" and child.tag == "param":
                child.tag = "set_parameter"

            # run recursively until we have converted the entire tree
            self.convert(root=child, path=path)

def _convert_param(parameter_name: str) -> str:
    if parameter_name[0] == "~":
        parameter_name = parameter_name[1:]

    parameter_name = parameter_name.replace("/", ".")
    if parameter_name[0] == ".":
        parameter_name = parameter_name[1:]

    return parameter_name

def _evaluate_conditional(condition_str, args):
    """
    Helper to evaluate if and unless
    """
    comparison_pattern = r'\$\(eval\s*arg\((?:[\'"]?)(\w+)(?:[\'"]?)\)\s*([=!<>]+)\s*(?:[\'"]?)(.*?)(?:[\'"]?)\)'
    match = re.search(comparison_pattern, condition_str)

    if match:
        arg_name, op_symbol, expected_val = match.groups()
        if arg_name not in args:
            return False

        actual_val = args[arg_name]

        try:
            expected_val = type(actual_val)(expected_val)
        except (ValueError, TypeError):
            pass

        operator_map = {
            "==": lambda a, b: a == b,
            "!=": lambda a, b: a != b,
            "<": lambda a, b: a < b,
            "<=": lambda a, b: a <= b,
            ">": lambda a, b: a > b,
            ">=": lambda a, b: a >= b,
        }
        return operator_map.get(op_symbol, lambda a, b: False)(actual_val, expected_val)

    # If not a comparison, check for a simple variable reference like $(arg ...)
    variable_pattern = r"\$\((?:arg|var)\s*(\w+)\)"
    match = re.search(variable_pattern, condition_str)

    value_to_check = args.get(match.group(1)) if match else condition_str

    return str(value_to_check).lower() in ["true", "1"]
