"""
ROS2 version
"""

import time
import ast

import armw.globals
import armw.ros2.wait_for_message
from armw.ros2.exceptions import ROSException
from armw.ros2.executors import MultiThreadedExecutor
import rclpy
from armw.base_node import BaseNode
from rclpy.callback_groups import MutuallyExclusiveCallbackGroup, ReentrantCallbackGroup
from rclpy.node import Node
from rclpy.qos import QoSProfile, QoSDurabilityPolicy
from ros2topic.api import get_msg_class
from deprecated import deprecated

executor: MultiThreadedExecutor = None


class ServiceProxy(object):
    """
    Works either like a ros2 service or a ros1 service

    If blocking, we wait on future and only return once its been fufilled. If
    not blocking, we return a future and leave timeouts up to the user.
    """

    def __init__(self, service_name, service_type, callback_group, blocking):
        self._client = armw.NODE().create_client(
            service_type, service_name, callback_group=callback_group
        )
        self.service_class = service_type.Request

        self.blocking = blocking

    def __call__(self, *args, **kwargs):
        global executor

        if len(args) > 0:
            req = args[0]
        else:
            req = self.service_class(**kwargs)

        future = self._client.call_async(req)
        if self.blocking:
            executor.might_spin_until_future_complete(future)
            return future.result()
        else:
            return future


class Ros2Node(BaseNode, Node):
    def __init__(self, node_name=None):
        # Run parent constructors
        global executor

        node_name = node_name if node_name else armw.globals.NODE_NAME

        Node.__init__(
            self,
            node_name,
            allow_undeclared_parameters=True,
            automatically_declare_parameters_from_overrides=True,
        )

        BaseNode.__init__(self)

        armw.globals.NODE = self
        armw.globals.NODE_NAME = node_name

        executor = MultiThreadedExecutor()
        executor.add_node(armw.globals.NODE)

        self.timer_callback_group = MutuallyExclusiveCallbackGroup()
        self.service_callback_group = ReentrantCallbackGroup()
        self.service_client_callback_group = ReentrantCallbackGroup()
        self.subscriber_callback_group = ReentrantCallbackGroup()
        self.publisher_callback_group = ReentrantCallbackGroup()

        # Need so do some silly stuff to make services work
        self.service_callbacks = {}

    def shutdown(self):
        self.destroy_node()

    def log(self, level: int, message: str):
        if level == -1:
            self.get_logger().debug(message)
        elif level == 0:
            self.get_logger().info(message)
        elif level == 1:
            self.get_logger().warn(message)
        elif level == 2:
            self.get_logger().error(message)
        elif level == 3:
            self.get_logger().fatal(message)

    def log_throttle(self, interval: int, level: int, message: str):
        if level == -1:
            self.get_logger().debug(message, throttle_duration_sec=interval)
        elif level == 0:
            self.get_logger().info(message, throttle_duration_sec=interval)
        elif level == 1:
            self.get_logger().warn(message, throttle_duration_sec=interval)
        elif level == 2:
            self.get_logger().error(message, throttle_duration_sec=interval)
        elif level == 3:
            self.get_logger().fatal(message, throttle_duration_sec=interval)

    def get_name(self):
        return Node.get_name(self)

    def get_namespace(self):
        return Node.get_namespace(self)

    def search_param(self, parameter_name):
        """
        This isn't really a thing in ROS2
        """

        return parameter_name

    def has_param(self, parameter_name):
        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]
        parameter_name = parameter_name.replace("/", ".")

        return self.has_parameter(parameter_name)

    def get_param(self, parameter_name, default=None):
        """
        Returns the value for a given parameter name, or None if it can't find it
        """

        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]

        if len(parameter_name) == 0:
            return default

        parameter_name = parameter_name.replace("/", ".")
        if parameter_name[0] == ".":
            parameter_name = parameter_name[1:]

        try:
            if default is not None and isinstance(default, dict):
                self.log_warn("Using get_param for dicts has been deprecated as of v2.0.0, this capability will be removed in the future, please call get_param_as_dict instead")
                return self.get_param_as_dict(parameter_name, default)
            else:
                param_value = self.get_parameter(parameter_name).get_parameter_value()

                if param_value.type == 0:  # Parameter not set
                    return default
                return rclpy.parameter.parameter_value_to_python(param_value)
        except Exception as e:
            self.log_error(f"Unable to read parameter {parameter_name}: {e}")
            return default

    def get_param_as_dict(self, parameter_name, default=None):
        """
        Returns a nested dictionary of parameters matching the given prefix.
        """
        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]

        if len(parameter_name) == 0:
            return default

        parameter_name = parameter_name.replace("/", ".")
        if parameter_name[0] == ".":
            parameter_name = parameter_name[1:]

        param_dict = self.get_parameters_by_prefix(parameter_name)

        # If no parameters match the prefix, return the default value.
        if not param_dict:
            return default

        return_value = {}

        for full_key, param_obj in param_dict.items():
            param_value = param_obj.get_parameter_value()
            raw_value = rclpy.parameter.parameter_value_to_python(param_value)

            key_parts = full_key.split('.')

            current_level = return_value
            for part in key_parts[:-1]:
                current_level = current_level.setdefault(part, {})

            current_level[key_parts[-1]] = raw_value

        return return_value

    def set_param(self, parameter_name, value):
        if parameter_name[0] == "~":
            parameter_name = parameter_name[1:]
        parameter_name = parameter_name.replace("/", ".")

        parameter = rclpy.Parameter(
            parameter_name, rclpy.Parameter.Type.from_parameter_value(value), value
        )
        self.set_parameters([parameter])

    def fix_topic_name(self, topic: str) -> str:
        """
        ros allows ~ at the start of a topic but requires ~ to be followed by /

        :param topic: topic to splice

        :return: spliced string
        """
        if topic.startswith("~") and not topic.startswith("~/"):
            return f"~/{topic[1:]}"
        return topic

    def get_topic_type(self, topic):
        return get_msg_class(self, topic, blocking=False)

    def publish(self, topic: str, message_type, queue_size=10, latch=False):
        if latch:
            qos = QoSProfile(depth=1, durability=QoSDurabilityPolicy.TRANSIENT_LOCAL)
        else:
            qos = queue_size

        return self.create_publisher(message_type, self.fix_topic_name(topic), qos, callback_group=self.publisher_callback_group)

    def subscribe(self, topic: str, message_type, callback, queue_size=10, *, latch=False):
        """
        Create a subscriber

        If you want to subscribe to a latched topic in ROS2, you need to set a transient local QoS.

        :param topic: topic to sub to
        :param message_type: type of message to sub to
        :param callback: callback function to call on message recv
        :param queue_size: how many messages to queue
        :param latch: sets the QoS to transient local if subbing to a latched topic

        """
        # This will work if there's a publisher up, but it's not recommended
        if message_type is None:
            self.log_warn("-" * 1000)
            self.log_warn(
                f"Subscribing to anymsg on topic {topic}.  In ROS2 this requires waiting for a publisher to exist"
            )
            message_type = get_msg_class(self, topic, blocking=True)

        qos = QoSProfile(depth=queue_size, durability=QoSDurabilityPolicy.TRANSIENT_LOCAL) if latch else queue_size

        return self.create_subscription(
            message_type, self.fix_topic_name(topic), callback, qos, callback_group=self.subscriber_callback_group
        )

    def create_service_server(self, name, service_type, callback):
        self.service_callbacks[name] = callback

        return self.create_service(
            service_type,
            self.fix_topic_name(name),
            lambda req, resp: self.service_callback(req, resp, name),
            callback_group=self.service_callback_group
        )

    def service_callback(self, request, response, name):
        response = self.service_callbacks[name](request)
        return response

    def create_service_client(self, name: str, service_type, persistent=False, blocking=True):
        # return self.create_client(service_type, name)
        return ServiceProxy(
            self.fix_topic_name(name), service_type, self.service_client_callback_group, blocking=blocking
        )

    def wait_for_message(self, topic_name, topic_type, timeout=None, *, latch=False):
        if timeout is None:
            time_to_wait = 100000  # Essentially forever
        else:
            time_to_wait = timeout

        self.log_debug(
            f"Waiting for message on {topic_name} for {time_to_wait} seconds"
        )

        if topic_type is None:
            from ros2topic.api import get_msg_class

            topic_type = get_msg_class(
                self, self.fix_topic_name(topic_name), blocking=True
            )

        qos = QoSProfile(depth=1, durability=QoSDurabilityPolicy.TRANSIENT_LOCAL) if latch else 1

        success, msg = armw.ros2.wait_for_message.wait_for_message(
            topic_type, self, self.fix_topic_name(topic_name), time_to_wait=time_to_wait, qos_profile=qos, callback_group=self.subscriber_callback_group
        )

        if success:
            self.log_debug(f"Got message on {topic_name}")
            return msg
        else:
            raise ROSException(f"Timed out waiting for message on {topic_name}")

    def wait_for_service(self, service_name, timeout=None):
        """
        This might not work correctly yet
        """
        if timeout is not None and timeout <= 0:
            raise ValueError("timeout needs to be greater than 0")

        start_time = time.time()
        end_time = start_time + timeout if timeout is not None else float("inf")

        service_name = self.fix_topic_name(service_name)

        # Resolve the name
        if service_name[0] != "/":
            service_name = self.get_namespace() + "/" + service_name
            service_name = armw.canonicalize_name(service_name)

        # Wait
        self.log_warn(f"Waiting for service {service_name}")
        while time.time() < end_time:
            for service in self.get_service_names_and_types():
                if service[0] == service_name:
                    self.log_warn(f"Got service {service_name}")
                    return
            time.sleep(0.1)


def init():
    """
    rclpy init
    """
    rclpy.init()


def spin(node, *args) -> None:
    """
    Implementation of rclpy's node::spin.

    Sets executor spinning property to ensure we are not spinning in multiple
    places. This fixes a very very obnoxious deadlock.
    """
    global executor

    with executor._spinning_lock:
        if executor._spinning:
            raise ValueError("You are trying to spin in multiple places")
        executor._spinning = True

    rclpy.spin(node, executor=executor)
    try:
        executor.add_node(node)
        while executor.context.ok():
            executor.spin_once()
    finally:
        executor.remove_node(node)

@deprecated(version='2.0.0', reason="This function is deprecated, please use init instead")
def init_node(node_name) -> None:
    rclpy.init()
    armw.globals.NODE_NAME = node_name


@deprecated(version='2.0.0', reason="This function is deprecated, please use init and spin directly instead")
def run_node(node_object):
    armw.globals.NODE = node_object
    try:
        spin(node_object)
    except KeyboardInterrupt:
        pass
    finally:
        rclpy.shutdown()


@deprecated(version='2.0.0', reason="This function is deprecated, please use init and spin directly instead")
def main(node_class, node_name) -> None:
    """
    Provides a simple interface to spin up a ros node while reducing some
    boilerplate

    :param node_class: Node to run
    :param node_name: Name of node to run

    :return: None
    """
    init()

    armw.globals.NODE_NAME = node_name

    try:
        spin(node_class())
    except KeyboardInterrupt:
        pass
    finally:
        rclpy.shutdown()
