import types

import rclpy
import rclpy.clock
import rclpy.duration
import builtin_interfaces
import builtin_interfaces.msg
import armw.globals

import time


def get_time():
    if armw.globals.NODE is not None:
        return armw.Time().now().to_sec()
    else:
        raise Exception("time is not initialized. Have you created a node?")


def sleep(duration: float):
    """
    works like ros1 sleep, preserves sim time usage
    """

    try:
        if duration < 0:

            armw.globals.NODE.get_logger().warn(
                f"Attempted to sleep for a negative duration - {duration}s."
            )
            return

        frequency = 1.0 / (duration.nanoseconds / 1_000_000_000.0)

        rate = armw.globals.NODE.create_rate(frequency)
        rate.sleep()
    except Exception as e:
        time.sleep(duration)


class Duration(rclpy.duration.Duration):
    def __init__(self, secs=0, nsecs=0):
        super().__init__(nanoseconds=secs * 1_000_000_000 + nsecs)
        self.to_nsec = types.MethodType(lambda self: self.nanoseconds, self)
        self.to_sec = types.MethodType(lambda self: self.nanoseconds / 1e9, self)
        self.is_zero = types.MethodType(lambda self: self.nanoseconds == 0, self)
        self.secs = secs
        self.nsecs = nsecs

    def __add__(self, other):
        if isinstance(other, Time):
            return armw.Time().from_sec(self.to_sec() + other.to_sec())
        elif isinstance(other, Duration):
            new_nanoseconds = self.to_nsec() + other.to_nsec()
            return Duration(nanoseconds=new_nanoseconds)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Time):
            return armw.Time().from_sec(self.to_sec() - other.to_sec())
        elif isinstance(other, Duration):
            new_nanoseconds = self.to_nsec() - other.to_nsec()
            return Duration(nsecs=new_nanoseconds)
        return NotImplemented

    @classmethod
    def from_sec(cls, secs):
        return Duration(int(secs), nsecs=int((secs - int(secs)) * 1_000_000_000))

    @classmethod
    def from_seconds(cls, secs):
        return Duration(int(secs), nsecs=int((secs - int(secs)) * 1_000_000_000))


class Time(builtin_interfaces.msg.Time):
    def __init__(self, secs=0, nsecs=0):
        super().__init__(sec=secs, nanosec=nsecs)

    def __add__(self, other):
        if isinstance(other, rclpy.duration.Duration) or isinstance(other, Duration):
            new_total_nanoseconds = self.to_nsec() + other.to_nsec()
            new_secs = new_total_nanoseconds // 1_000_000_000
            new_nsecs = new_total_nanoseconds % 1_000_000_000
            return Time(secs=new_secs, nsecs=new_nsecs)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, rclpy.time.Time) or isinstance(other, Time):
            new_nanoseconds_diff = self.to_nsec() - other.to_nsec()
            return Duration(nsecs=new_nanoseconds_diff)
        elif isinstance(other, rclpy.duration.Duration) or isinstance(other, Duration):
            new_total_nanoseconds = self.to_nsec() - other.to_nsec()
            new_secs = new_total_nanoseconds // 1_000_000_000
            new_nsecs = new_total_nanoseconds % 1_000_000_000
            return Time(secs=new_secs, nsecs=new_nsecs)
        return NotImplemented

    def __gt__(self, other):
        return self.to_sec() > other.to_sec()

    def __lt__(self, other):
        return self.to_sec() < other.to_sec()

    def __ge__(self, other):
        return self.to_sec() >= other.to_sec()

    def __le__(self, other):
        return self.to_sec() <= other.to_sec()

    def __eq__(self, other):
        return self.to_sec() == other.to_sec()

    def __ne__(self, other):
        return self.to_sec() != other.to_sec()

    def __hash__(self):
        return hash(self.to_sec())

    def to_sec(self):
        return self.sec + float(self.nanosec) / 1_000_000_000

    def to_nsec(self):
        return self.sec * 1_000_000_000 + self.nanosec

    @classmethod
    def from_sec(cls, secs: float):
        return Time(secs=int(secs), nsecs=int((secs - int(secs)) * 1_000_000_000))

    @classmethod
    def from_seconds(cls, secs: float):
        return Time().from_sec(secs=secs)

    @classmethod
    def now(cls):
        s = armw.globals.NODE.get_clock().now()

        total_nanoseconds = s.nanoseconds

        seconds = total_nanoseconds // 1_000_000_000
        fractional_nanoseconds = total_nanoseconds % 1_000_000_000

        return Time(secs=seconds, nsecs=fractional_nanoseconds)


class Rate(object):
    def __init__(self, hz):
        self._rate = armw.globals.NODE.create_rate(hz)

    def __del__(self):
        armw.globals.NODE.destroy_rate(self._rate)

    def sleep(self):
        self._rate.sleep()


if __name__ == "__main__":
    a = Time(3, 5)
    b = Duration(5, 6)

    print(a)
    print(b)

    print(a + b)
    print(b + a)
